<?php

/**
 * @file
 * Admin pages for Process Api.
 */

/**
 * Page callback that shows an overview of defined servers.
 */
function process_api_admin_servers() {
  $base_path = drupal_get_path('module', 'process_api') . '/';
  drupal_add_css($base_path . 'process_api.admin.css');
  drupal_add_js($base_path . 'process_api.admin.js');

  $servers = process_api_server_load_multiple(FALSE);

  // Show disabled servers after enabled ones.
  foreach ($servers as $id => $server) {
    if (!$server->enabled) {
      unset($servers[$id]);
      $servers[$id] = $server;
    }
  }

  $rows = array();
  $t_enabled['data'] = array(
    '#theme' => 'image',
    '#path' => $base_path . 'enabled.png',
    '#alt' => t('enabled'),
    '#title' => t('enabled'),
  );
  $t_enabled['class'] = array('process-api-status');
  $t_disabled['data'] = array(
    '#theme' => 'image',
    '#path' => $base_path . 'disabled.png',
    '#alt' => t('disabled'),
    '#title' => t('disabled'),
  );
  $t_disabled['class'] = array('process-api-status');
  $t_enable = t('enable');
  $t_disable = t('disable');
  $t_edit = t('edit');
  $t_delete = t('delete');
  $pre_server = 'admin/config/system/process_api/server';
  $enable = '/enable';
  $disable = '/disable';
  $edit = '/edit';
  $delete = '/delete';
  $edit_link_options['attributes']['class'][] = 'process-api-edit-menu-toggle';
  foreach ($servers as $server) {
    $url = $pre_server . '/' . $server->machine_name;
    $row = array();
    $row[] = $server->enabled ? $t_enabled : $t_disabled;
    $row[] = l($server->name, $url);
    $row[] = $server->enabled ? l($t_disable, $url . $disable) : l($t_enable, $url . $enable);
    $row[] = l($t_edit, $url . $edit);
    $row[] = l($t_delete, $url . $delete); // see threads ENTITY_IN_DB ... some weird stuff TODO
    $rows[] = $row;
    
  }

  $header = array();
  $header[] = t('Status');
  $header[] = t('Name');
  $header[] = array('data' => t('Operations'), 'colspan' => 3);

  return array(
    '#theme' => 'table',
    '#header' => $header,
    '#rows' => $rows,
    '#empty' => t('There are no process servers defined yet.'),
  );
}

/**
 * Page callback that shows an overview of defined threads.
 */
function process_api_admin_threads($form, &$form_state) {
  $base_path = drupal_get_path('module', 'process_api') . '/';
  
  $form['#tree'] = TRUE;

  $show_config_status = FALSE;
  $threads=process_api_thread_load_multiple(FALSE, array(), FALSE, TRUE);
  foreach ($threads as $thread) {
    if (!$show_config_status && $thread->status != ENTITY_IN_DB) {
      $show_config_status = TRUE;
    }
  }
  
  $form['#show_config_status'] = $show_config_status;
  $pre_server = 'admin/config/system/process_api/server';
  $pre_thread = 'admin/config/system/process_api/thread';
  
  foreach ($threads as $thread) {
    $server=$thread->server();
    $form[$thread->machine_name]['#thread'] = $thread;
    if ($thread->enabled) {
      $form[$thread->machine_name]['status']['#markup'] = theme(
          'image',
          array(
            'path' => $base_path . 'enabled.png',
            'alt' => t('enabled'),
            'title' => t('enabled'))
        );
    }
    else {
      $form[$thread->machine_name]['status']['#markup'] = theme(
          'image',
          array(
            'path' => $base_path . 'disabled.png',
            'alt' => t('disabled'),
            'title' => t('disabled'))
        );
    }
    if ($show_config_status) {
      $form[$thread->machine_name]['entity_status']['#markup'] = theme('entity_status', array('status' => $thread->status));
    }
    $form[$thread->machine_name]['server'] = array(
      '#type' => 'link',
      '#title' => check_plain($server->name),
      '#href' => "$pre_server/$server->machine_name",
    );
    $form[$thread->machine_name]['thread'] = array(
      '#type' => 'link',
      '#title' => check_plain($thread->name),
      '#href' => "$pre_thread/$thread->machine_name",
    );
    $form[$thread->machine_name]['toggle'] = array(
      '#type' => 'link',
      '#title' => $thread->enabled ? t('Disable') : t('Enable'),
      '#href' => "$pre_thread/$thread->machine_name" . ($thread->enabled?'/disable':'/enable'),
    );
    $form[$thread->machine_name]['edit'] = array(
      '#type' => 'link',
      '#title' => t('Edit'),
      '#href' => "$pre_thread/$thread->machine_name/edit",
      '#attributes' => array('class' => array('process-api-edit-menu-toggle')),
      '#suffix' =>
        '<div class="process-api-edit-menu collapsed">' .
        theme('links', array('links' => menu_contextual_links('process-api-thread', $pre_thread, array($thread->machine_name)))) .
        '</div>',
    );
    $form[$thread->machine_name]['delete'] = array(
      '#type' => 'link',
      '#title' => t('Delete'),
      '#href' => "$pre_thread/$thread->machine_name/delete",
    );
    $form[$thread->machine_name]['weight'] = array(
      '#type' => 'weight',
      '#title' => t('Weight for @title', array('@title' => $thread->name)),
      '#title_display' => 'invisible',
      '#delta' => 10,
      '#default_value' => $thread->weight,
    );
  }

  // Only make this form include a submit button and weight if more than one
  // thread exists.
  if (count($threads) > 1) {
    $form['actions'] = array('#type' => 'actions');
    $form['actions']['submit'] = array('#type' => 'submit', '#value' => t('Save'));
  }
  elseif (isset($thread)) {
    unset($form[$thread->machine_name]['weight']);
  }
  return $form;
}

/**
 * Submit handler for threads overview. Updates changed vocabulary weights.
 *
 * @see taxonomy_overview_vocabularies()
 */
function process_api_admin_threads_submit($form, &$form_state) {
  print_r($form_state['values']);
  foreach ($form_state['values'] as $machine_name => $thread) {
    if (!isset($form[$machine_name]['#thread'])) continue;
    if ($form[$machine_name]['#thread']->weight != $form_state['values'][$machine_name]['weight']) {
      $form[$machine_name]['#thread']->snapshot('weight');
      $form[$machine_name]['#thread']->weight = $form_state['values'][$machine_name]['weight'];
      $form[$machine_name]['#thread']->update();
    }
  }
  drupal_set_message(t('The configuration options have been saved.'));
}

/**
 * Returns HTML for the threads overview form.
 *
 * @param $variables
 *   An associative array containing:
 *   - form: A render element representing the form.
 *
 * @see process_api_admin_threads()
 * @ingroup themeable
 */
function theme_process_api_admin_threads(&$variables) {
  $form = $variables['form'];

  $base_path = drupal_get_path('module', 'process_api') . '/';
  drupal_add_css($base_path . 'process_api.admin.css');
  drupal_add_js($base_path . 'process_api.admin.js');
  
  $show_config_status = $form['#show_config_status'];
  
  $rows = array();

  foreach (element_children($form) as $key) {
    if (isset($form[$key]['thread'])) {
      $thread = &$form[$key];

      $row = array();
      $row[] = drupal_render($thread['status']);
      if ($show_config_status) {
        $row[] = drupal_render($thread['entity_status']);
      }
      $row[] = drupal_render($thread['server']);
      $row[] = drupal_render($thread['thread']);
      $row[] = drupal_render($thread['toggle']);
      $row[] = drupal_render($thread['edit']);
      $row[] = drupal_render($thread['delete']);

      if (isset($thread['weight'])) {
        $thread['weight']['#attributes']['class'] = array('thread-weight');
        $row[] = drupal_render($thread['weight']);
      }
      
      $rows[] = array('data' => $row, 'class' => array('draggable'));
    }
  }

  $header = array();
  $header[] = t('Status');
  if ($show_config_status) {
    $header[] = t('Configuration');
  }
  $header[] = t('Server'); 
  $header[] = t('Thread');
  $header[] = array('data' => t('Operations'), 'colspan' => 3);
  if (isset($form['actions'])) {
    $header[] = t('Weight');
    drupal_add_tabledrag('threads-order', 'order', 'sibling', 'thread-weight');
  }
  return theme('table', array('header' => $header, 'rows' => $rows, 'empty' => t('There are no process threads defined yet.'), 'attributes' => array('id' => 'threads-order'))) . drupal_render_children($form);  
}

/**
 * @param Entity $entity
 *   The thread or server for which a link should be generated.
 *
 * @return string
 *   A link to a delete form for the entity, if applicable.
 */
function _process_api_admin_delete_link(Entity $entity) {
  // Delete link only makes sense if entity is in the database (custom or overridden).
  if ($entity->status & ENTITY_IN_DB) {
    $type = $entity instanceof ProcessApiServer ? 'server' : 'thread';
    $url = 'admin/config/system/process_api/' . $type . '/' . $entity->machine_name . '/delete';
    $title = $entity->status == ENTITY_IN_DB ? t('delete') : t('revert');
    return l($title, $url);
  }
  return '';
}

/**
 * Form callback showing a form for adding a server.
 */
function process_api_admin_add_server(array $form, array &$form_state) {
  drupal_set_title(t('Add server'));

  $class = empty($form_state['values']['class']) ? '' : $form_state['values']['class'];
  $server = $form_state['server'] = new ProcessApiServer();

  if (empty($form_state['storage']['step_one'])) {
    $form['name'] = array(
      '#type' => 'textfield',
      '#title' => t('Server name'),
      '#description' => t('Enter the displayed name for the new server.'),
      '#maxlength' => 50,
      '#required' => TRUE,
    );

    $form['machine_name'] = array(
      '#type' => 'machine_name',
      '#maxlength' => 50,
      '#machine_name' => array(
        'exists' => 'process_api_server_load',
      ),
    );

    $form['enabled'] = array(
      '#type' => 'checkbox',
      '#title' => t('Enabled'),
      '#description' => t('Select if the new server will be enabled after creation.'),
      '#default_value' => TRUE,
    );
    $form['description'] = array(
      '#type' => 'textarea',
      '#title' => t('Server description'),
      '#description' => t('Enter a description for the new server.'),
    );
    $form['class'] = array(
      '#type' => 'select',
      '#title' => t('Service class'),
      '#description' => t('Choose a service class to use for this server.'),
      '#options' => array('' => '< ' . t('Choose a service class') . ' >'),
      '#required' => TRUE,
      '#default_value' => $class,
      '#ajax' => array(
        'callback' => 'process_api_admin_add_server_ajax_callback',
        'wrapper' => 'process-api-class-options',
      ),
    );
    
    foreach (process_api_get_service_info() as $_class => $info) {
      $form['class']['#options'][$_class] = $info['name'];
    }
    
  }
  elseif (!$class) {
    $class = $form_state['storage']['step_one']['class'];
  }

  if ($class) {
    $info=process_api_get_service_info($class);
    //TODO rework the =$class
    $server->class=$class;
    $service = NULL;
    if (class_exists($info['class'])) {
      $service = new $info['class']($server);
    }
    if (!($service instanceof ProcessApiServiceInterface)) {
      watchdog('process_api', 'Service class !id specifies an illegal class: !class', array('!id' => $id, '!class' => $info['class']), WATCHDOG_ERROR);
      continue;
    }
    $service_form = $server->configurationForm($form, $form_state);
    $form['options']['form'] = $service_form ? $service_form : array('#markup' => t('There are no server configuration options for this service class.'));
    $form['options']['#type'] = 'fieldset';
    $form['options']['#tree'] = TRUE;
    $form['options']['#collapsible'] = TRUE;
    $form['options']['#title'] = $info['name'];
    $form['options']['#description'] = $info['description'];
    // be sure of configured class
    $form['options']['class']['#type'] = 'value';
    $form['options']['class']['#value'] = $class;
  }
  $form['options']['#prefix'] = '<div id="process-api-class-options">';
  $form['options']['#suffix'] = '</div>';

  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Create server'),
  );

  return $form;
}

/**
 * AJAX callback that just returns the "options" array of the already built form
 * array.
 */
function process_api_admin_add_server_ajax_callback(array $form, array &$form_state) {
  return $form['options'];
}

/**
 * Form validation callback for adding a server.
 *
 * Validates the machine name and calls the service class' validation handler.
 */
function process_api_admin_add_server_validate(array $form, array &$form_state) {

  $server = $form_state['server'];
  if (!empty($form_state['values']['machine_name'])) {
    $name = $form_state['values']['machine_name'];
    if (is_numeric($name)) {
      form_set_error('machine_name', t('The machine name must not be a pure number.'));
    }
  }
  
  if (empty($form_state['values']['options']['class'])) {
    // why empty ?
    return;
  }

  $class = $form_state['values']['options']['class'];
  $info = process_api_get_service_info($class);
  $service = NULL;
  if (class_exists($info['class'])) {
    $service = new $info['class']($server);
  }
  if (!($service instanceof ProcessApiServiceInterface)) {
    form_set_error('class', t('There seems to be something wrong with the selected service class.'));
    return;
  }
  // TODO very weird code here...
  // $form_state['values']['options']['service'] = $service;
  $server->class=$class;
  $server->configurationFormValidate(
    $form,
    isset($form_state['values']['options']['form'])?$form_state['values']['options']['form']:array(),
    $form_state);
}

/**
 * Form submit callback for adding a server.
 */
function process_api_admin_add_server_submit(array $form, array &$form_state) {
  form_state_values_clean($form_state);
  $values = $form_state['values'];
  $server = $form_state['server'];

  if (!empty($form_state['storage']['step_one'])) {
    $values += $form_state['storage']['step_one'];
    unset($form_state['storage']);
  }

  if (empty($values['options']) || ($values['class'] != $values['options']['class'])) {
    unset($values['options']);
    $form_state['storage']['step_one'] = $values;
    //$form_state['rebuild'] = TRUE;
    drupal_set_message(t('Please configure the used service.'));
    return;
  }

  $options = isset($values['options']['form']) ? $values['options']['form'] : array();
  unset($values['options']); //options are set in configurationFormValidate();
  //$form_state['server']  = $server = new ProcessApiServer($values); TODO remove
  $server->setValues($values);
  $server->configurationFormSubmit($form, $options, $form_state);
  $server->save();
  $form_state['redirect'] = 'admin/config/system/process_api/server/' . $server->machine_name;
  drupal_set_message(t('The server was successfully created.'));
}

/**
 * Title callback for viewing or editing a server or thread.
 */
function process_api_admin_item_title($object) {
  return $object->name;
}

/**
 * Displays a server's details.
 *
 * @param ProcessApiServer $server
 *   The server to display.
 * @param $action
 *   One of 'enable', 'disable', 'delete'; or NULL if the server is only viewed.
 */
function process_api_admin_server_view(ProcessApiServer $server, $action = NULL) {
  if (!empty($action)) {
    if ($action == 'enable') {
      if ($server->enable()) {
        drupal_set_message(t('The server was successfully enabled.'));
      }
      else {
        drupal_set_message(t('The server could not be enabled. Check the logs for details.'), 'error');
      }
      drupal_goto('admin/config/system/process_api/server/' . $server->machine_name);
    }
    else {
      $ret = drupal_get_form('process_api_admin_confirm', 'server', $action, $server);
      if ($ret) {
        return $ret;
      }
    }
  }

  drupal_set_title(check_plain(process_api_admin_item_title($server)));
  $class = process_api_get_service_info($server->class);
  $options = $server->viewSettings();
  return array(
    '#theme' => 'process_api_server',
    '#id' => $server->id,
    '#name' => $server->name,
    '#machine_name' => $server->machine_name,
    '#description' => $server->description,
    '#enabled' => $server->enabled,
    '#class_name' => $class['name'],
    '#class_description' => $class['description'],
    '#options' => $options,
    '#status' => $server->status,
  );
}

/**
 * Theme function for displaying a server.
 *
 * @param array $variables
 *   An associative array containing:
 *   - id: The server's id.
 *   - name: The server's name.
 *   - machine_name: The server's machine name.
 *   - description: The server's description.
 *   - enabled: Boolean indicating whether the server is enabled.
 *   - class_name: The used service class' display name.
 *   - class_description: The used service class' description.
 *   - options: An HTML string or render array containing information about the
 *     server's service-specific settings.
 *   - status: The entity configuration status (in database, in code, etc.).
 */
function theme_process_api_server(array $variables) {
  extract($variables);
  $output = '';

  $output .= '<h3>' . check_plain($name) . '</h3>' . "\n";

  $output .= '<dl>' . "\n";

  $output .= '<dt>' . t('Status') . '</dt>' . "\n";
  $output .= '<dd>';
  if ($enabled) {
    $output .= t('enabled (!disable_link)', array('!disable_link' => l(t('disable'), 'admin/config/system/process_api/server/' . $machine_name . '/disable')));
  }
  else {
    $output .= t('disabled (!enable_link)', array('!enable_link' => l(t('enable'), 'admin/config/system/process_api/server/' . $machine_name . '/enable')));
  }
  $output .= '</dd>' . "\n";

  $output .= '<dt>' . t('Machine name') . '</dt>' . "\n";
  $output .= '<dd>' . check_plain($machine_name) . '</dd>' . "\n";

  if (!empty($description)) {
    $output .= '<dt>' . t('Description') . '</dt>' . "\n";
    $output .= '<dd>' . nl2br(check_plain($description)) . '</dd>' . "\n";
  }

  if (!empty($class_name)) {
    $output .= '<dt>' . t('Service class') . '</dt>' . "\n";
    $output .= '<dd><em>' . check_plain($class_name) . '</em>';
    if (!empty($class_description)) {
      $output .= '<p class="description">' . $class_description . '</p>';
    }
    $output .= '</dd>' . "\n";
  }

  if (!empty($options)) {
    $output .= '<dt>' . t('Service options') . '</dt>' . "\n";
    $output .= '<dd>' . "\n";
    $output .= render($options);
    $output .= '</dd>' . "\n";
  }

  $output .= '<dt>' . t('Configuration status') . '</dt>' . "\n";
  $output .= '<dd>' . "\n";
  $output .= theme('entity_status', array('status' => $status));
  $output .= '</dd>' . "\n";

  $output .= '</dl>';

  return $output;
}

/**
 * Edit a server's settings.
 *
 * @param ProcessApiServer $server
 *   The server to edit.
 */
function process_api_admin_server_edit(array $form, array &$form_state, ProcessApiServer $server) {
  $form_state['server'] = $server;

  $form['name'] = array(
    '#type' => 'textfield',
    '#title' => t('Server name'),
    '#description' => t('Enter the displayed name for the  server.'),
    '#maxlength' => 50,
    '#default_value' => $server->name,
    '#required' => TRUE,
  );
  $form['description'] = array(
    '#type' => 'textarea',
    '#title' => t('Server description'),
    '#description' => t('Enter a description for the new server.'),
    '#default_value' => $server->description,
  );

  $class = process_api_get_service_info($server->class);

  $form['options']['form'] = $server->configurationForm($form, $form_state);
  $form['options']['#type'] = 'fieldset';
  $form['options']['#tree'] = TRUE;
  $form['options']['#collapsible'] = TRUE;
  $form['options']['#title'] = $class['name'];
  $form['options']['#description'] = $class['description'];

  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Save settings'),
  );

  return $form;
}

/**
 * Validation function for process_api_admin_server_edit.
 */
function process_api_admin_server_edit_validate(array $form, array &$form_state) {
  $form_state['server']->configurationFormValidate($form, $form_state['values']['options']['form'], $form_state);
}

/**
 * Submit function for process_api_admin_server_edit.
 */
function process_api_admin_server_edit_submit(array $form, array &$form_state) {
  form_state_values_clean($form_state);
  $values = $form_state['values'];

  $server = $form_state['server'];
  if (isset($values['options'])) {
    $server->configurationFormSubmit($form, $values['options']['form'], $form_state);
  }
  unset($values['options']);

  $ret = $server->update($values);
  $form_state['redirect'] = 'admin/config/system/process_api/server/' . $server->machine_name;
  if ($ret) {
    drupal_set_message(t('The process server was successfully edited.'));
  }
  else {
    drupal_set_message(t('No values were changed.'));
  }
}

/**
 * Form callback showing a form for adding a thread.
 */
function process_api_admin_add_thread(array $form, array &$form_state) {
  drupal_set_title(t('Add thread'));

  $class = empty($form_state['values']['class']) ? '' : $form_state['values']['class'];
  $thread = $form_state['thread'] = new ProcessApiThread();
  
  $form['#attached']['css'][] = drupal_get_path('module', 'process_api') . '/process_api.admin.css';
  $form['name'] = array(
    '#type' => 'textfield',
    '#title' => t('Thread name'),
    '#maxlength' => 50,
    '#required' => TRUE,
  );

  $form['machine_name'] = array(
    '#type' => 'machine_name',
    '#maxlength' => 50,
    '#machine_name' => array(
      'exists' => 'process_api_thread_load',
    ),
  );

  $form['enabled'] = array(
    '#type' => 'checkbox',
    '#title' => t('Enabled'),
    '#description' => t('This will only take effect if the selected server is also enabled.'),
    '#default_value' => TRUE,
  );
  $form['description'] = array(
    '#type' => 'textarea',
    '#title' => t('Thread description'),
  );
  $form['server'] = array(
    '#type' => 'select',
    '#title' => t('Server'),
    '#description' => t('Select the server this thread should reside on. ' .
      'This setting cannot be changed afterwards.'),
    '#default_value' => '',
    '#required' => TRUE,
    '#options' => array('' => t('< No server >')),
    '#ajax' => array(
      'callback' => 'process_api_admin_add_thread_ajax_callback',
      'wrapper' => 'process-api-class-options',
    ),
  );
  $servers = process_api_server_load_multiple(FALSE);
  // List enabled servers first.
  foreach ($servers as $server) {
    if ($server->enabled) {
      $form['server']['#options'][$server->machine_name] = $server->name;
    }
  }
  foreach ($servers as $server) {
    if (!$server->enabled) {
      $form['server']['#options'][$server->machine_name] = t('@server_name (disabled)', array('@server_name' => $server->name));
    }
  }

  $form['server_dependent']['#prefix'] = '<div id="process-api-class-options">';
  $form['server_dependent']['#suffix'] = '</div>';
  $server_id = $thread->server;
  if (isset($form_state['values']['server'])&&$form_state['values']['server']) {
    $server_id = $form_state['values']['server'];
  }
  if ($server_id) {
    $class = process_api_get_service_info($thread->server($server_id)->class);
    
    $form['server_dependent']['#type'] = 'fieldset';
    $form['server_dependent']['#collapsible'] = FALSE;
    $form['server_dependent']['#title'] = $class['name'];
    $form['server_dependent']['#description'] = $class['description'];    
    
    $form['server_dependent']['entity_type'] = array(
      '#type' => 'select',
      '#title' => t('Entity type'),
      '#description' => t('Select the type of entity that will be processed in this thread. ' .
      'This setting cannot be changed afterwards.'),
      '#options' => array(),
    );
    foreach ($thread->server($server_id)->getSupportedEntityTypes() as $name => $entity) {
      if (entity_get_property_info($name)) {
        $form['server_dependent']['entity_type']['#options'][$name] = $entity['label'];
      }
    }
    
    $options_form = $thread->configurationForm($form, $form_state, $server_id);
    $form['server_dependent']['options']['form'] = $options_form ? $options_form : array('#markup' => t('There are no thread configuration options for this service class.'));
    $form['server_dependent']['options']['#tree'] = TRUE;
  }
  
  $form['cron_limit'] = array(
    '#type' => 'textfield',
    '#title' => t('Cron limit'),
    '#description' => t('Set how many items will be processed at most during each run of cron. ' .
        '"0" means that no items will be processed by cron for this thread, "-1" means that cron should thread all items at once.'),
    '#default_value' => PROCESS_API_DEFAULT_CRON_LIMIT,
    '#size' => 4,
    '#attributes' => array('class' => array('process-api-cron-limit')),
  );
  
  $form['live'] = array(
    '#type' => 'select',
    '#title' => t('Try process "live"'),
    '#options' => array('live' => t('Yes'), '' => t('No')),
    '#description' => t('Will try to process entities in live on insert and update. This may cost overhead.'),
    '#default_value' => '',
  );
  
  $form['process_once'] = array(
      '#type' => 'select',
      '#title' => t('Process only once'),
      '#options' => array('process_once' => t('Yes'), '' => t('No')),
      '#description' => t('Will try to process entities only once at insert.'),
      '#default_value' => '',
  );
  
  $form['cascade_update'] = array(
      '#type' => 'select',
      '#title' => t('Cascade update'),
      '#options' => array('cascade_update' => t('Yes'), '' => t('No')),
      '#description' => t('Entity updates in this thread will trigger other threads.'),
      '#default_value' => '',
  );

  $form['retroactive_process'] = array(
      '#type' => 'select',
      '#title' => t('Retroactive process'),
      '#options' => array('retroactive_process' => t('Yes'), '' => t('No')),
      '#description' => t('Process existing entities.'),
      '#default_value' => '',
  );
  
  $form['track_disabled'] = array(
      '#type' => 'select',
      '#title' => t('Keep track when disabled'),
      '#options' => array('track_disabled' => t('Yes'), '' => t('No')),
      '#description' => t('Keep track of entities modifications even if thread is disabled.'),
      '#default_value' => '',
  );
  
  $form['purge_range'] = array(
      '#type' => 'textfield',
      '#title' => t('Purge range'),
      '#description' => t('For non retroactive process, you can define a purge of the Process Api tables. ' .
          'The purge range is in days. Items older than this range will be purged. ' .
          '"-1" means that purge is disabled.'),
      '#default_value' => PROCESS_API_DEFAULT_PURGE_RANGE,
      '#size' => 4,
      '#attributes' => array('class' => array('process-api-purge-range')),
  );
  
  $form['node_operations'] = array(
      '#type' => 'fieldset',
      '#title' => t('Node operations'),
      '#description' => t('Live trigger Process API in bulk node operations.'),
      '#tree' => TRUE,
  );

  foreach (module_invoke_all('node_operations') as $operation => $array) {
    $form['node_operations'][$operation] = array(
      '#type' => 'select',
      '#title' => $array['label'],
      '#options' => array('live' => t('Yes'), '' => t('No')),
      '#default_value' => 'live',
    );
  }
  
  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Create thread'),
  );

  return $form;
}

/**
 * AJAX callback that just returns the "options" array of the already built form
 * array.
 */
function process_api_admin_add_thread_ajax_callback(array $form, array &$form_state) {
  return $form['server_dependent'];
}

/**
 * Validation callback for process_api_admin_add_thread.
 */
function process_api_admin_add_thread_validate(array $form, array &$form_state) {
  $name = $form_state['values']['machine_name'];
  if (is_numeric($name)) {
    form_set_error('machine_name', t('The machine name must not be a pure number.'));
  }
  $thread = $form_state['thread'];
  
  $cron_limit = $form_state['values']['cron_limit'];
  if ($cron_limit != '' . ((int) $cron_limit)) {
    // We don't enforce stricter rules and treat all negative values as -1.
    form_set_error('cron_limit', t('The cron limit must be an integer.'));
  }

  if (isset($form_state['values']['server'])&&$form_state['values']['server']) {
    // TODO server not required ?
    $thread->configurationFormValidate(
      $form,
      isset($form_state['values']['options']['form']) ? $form_state['values']['options']['form'] : array(),
      $form_state,
      $form_state['values']['server']);
  }
}

/**
 * Submit callback for process_api_admin_add_thread.
 */
function process_api_admin_add_thread_submit(array $form, array &$form_state) {
  form_state_values_clean($form_state);
  $values = $form_state['values'];
  
  $values['options']['cron_limit'] = $values['cron_limit'];
  $values['options']['live'] = $values['live'];
  $values['options']['process_once'] = $values['process_once'];
  $values['options']['cascade_update'] = $values['cascade_update'];
  $values['options']['retroactive_process'] = $values['retroactive_process'];
  $values['options']['track_disabled'] = $values['track_disabled'];
  $values['options']['purge_range'] = $values['purge_range'];
  $values['options']['node_operations'] = $values['node_operations'];
  $options = isset($values['options']['form']) ? $values['options']['form'] : array();

  unset($values['options']['form']); // should be set by configurationFormSubmit()
  unset($values['cron_limit']);
  unset($values['live']);
  unset($values['process_once']);
  unset($values['cascade_update']);
  unset($values['retroactive_process']);
  unset($values['track_disabled']);
  unset($values['purge_range']);
  unset($values['node_operations']);

  $thread = $form_state['thread'];
  $thread->is_new = TRUE;
  
  $thread->setValues($values);
  // server should be set now
  $thread->configurationFormSubmit($form, $options, $form_state);
  // Trying to create an enabled thread on a disabled server is handled elsewhere
  $thread->save();

  drupal_set_message(t('The thread was successfully created. Please set up its workflow now.'));
  $form_state['redirect'] = 'admin/config/system/process_api/thread/' . $values['machine_name'] . '/workflow';
}

/**
 * Displays a thread' details.
 *
 * @param ProcessApiThread $thread
 *   The thread to display.
 */
function process_api_admin_thread_view(ProcessApiThread $thread = NULL, $action = NULL) {
  if (empty($thread)) {
    return MENU_NOT_FOUND;
  }

  if (!empty($action)) {
    if ($action == 'enable') {
      if ($thread->enable()) {
        drupal_set_message(t('The thread was successfully enabled.'));
      }
      else {
        drupal_set_message(t('The thread could not be enabled. Check the logs for details.'), 'error');
      }
      drupal_goto('admin/config/system/process_api/thread/' . $thread->machine_name);
    }
    else {
      $ret = drupal_get_form('process_api_admin_confirm', 'thread', $action, $thread);
      if ($ret) {
        return $ret;
      }
    }
  }

  $ret = array(
    '#theme' => 'process_api_thread',
    '#id' => $thread->id,
    '#name' => $thread->name,
    '#machine_name' => $thread->machine_name,
    '#description' => $thread->description,
    '#entity_type' => $thread->entity_type,
    '#enabled' => $thread->enabled,
    '#server' => $thread->server(),
    '#standard_options' => $thread->options,
    '#options' => $thread->viewSettings(),
    '#status' => $thread->status,
  );

  return $ret;
}

/**
 * Theme function for displaying a thread.
 *
 * @param array $variables
 *   An associative array containing:
 *   - id: The thread's id.
 *   - name: The thread' name.
 *   - machine_name: The thread' machine name.
 *   - description: The thread' description.
 *   - entity_type: The type of entities stored in this thread.
 *   - enabled: Boolean indicating whether the thread is enabled.
 *   - server: The server this thread currently rests on, if any.
 *   - options: The thread' options, like cron limit.
 *   - processed_items: The number of items already processed in their latest
 *     version on this thread.
 *   - total_items: The total number of items that have to be processed for this
 *     thread.
 *   - status: The entity configuration status (in database, in code, etc.).
 */
function theme_process_api_thread(array $variables) {
  extract($variables);
  $output = '';

  $output .= '<h3>' . check_plain($name) . '</h3>' . "\n";

  $output .= '<dl>' . "\n";

  $output .= '<dt>' . t('Status') . '</dt>' . "\n";
  $output .= '<dd>';
  if ($enabled) {
    $output .= t('enabled (!disable_link)', array('!disable_link' => l(t('disable'), 'admin/config/system/process_api/thread/' . $machine_name . '/disable')));
  }
  elseif ($server && $server->enabled) {
    $output .= t('disabled (!enable_link)', array('!enable_link' => l(t('enable'), 'admin/config/system/process_api/thread/' . $machine_name . '/enable')));
  }
  else {
    $output .= t('disabled');
  }
  $output .= '</dd>' . "\n";

  $output .= '<dt>' . t('Machine name') . '</dt>' . "\n";
  $output .= '<dd>' . check_plain($machine_name) . '</dd>' . "\n";

  $output .= '<dt>' . t('Entity type') . '</dt>' . "\n";
  $type = entity_get_info($entity_type);
  $type = $type['label'];
  $output .= '<dd>' . check_plain($type) . '</dd>' . "\n";

  if (!empty($description)) {
    $output .= '<dt>' . t('Description') . '</dt>' . "\n";
    $output .= '<dd>' . nl2br(check_plain($description)) . '</dd>' . "\n";
  }

  if (!empty($server)) {
    $output .= '<dt>' . t('Server') . '</dt>' . "\n";
    $output .= '<dd>' . l($server->name, 'admin/config/system/process_api/server/' . $server->machine_name);
    if (!empty($server->description)) {
      $output .= '<p class="description">' . nl2br(check_plain($server->description)) . '</p>';
    }
    $output .= '</dd>' . "\n";
  }

  if (!empty($standard_options)) {
    $output .= '<dt>' . t('Standard thread options') . '</dt>' . "\n";
    $output .= '<dd><dl>' . "\n";
    $output .= '<dt>' . t('Cron limit') . '</dt>' . "\n";
    if (empty($standard_options['cron_limit'])) {
      $output .= '<dd>' . t("Don't thread during cron runs") . '</dd>' . "\n";
    }
    elseif ($standard_options['cron_limit'] < 0) {
      $output .= '<dd>' . t('Unlimited') . '</dd>' . "\n";
    }
    else {
      $output .= '<dd>' . format_plural($standard_options['cron_limit'], '1 item per cron run.', '@count items per cron run.') . '</dd>' . "\n";
    }

    $output .= '<dt>' . t('Try process live: %live', array('%live' => t(isset($standard_options['live'])&&$standard_options['live']?'yes':'no'))) . '</dt>' . "\n";
    $output .= '<dt>' . t('Process only once: %process_once', array('%process_once' => t(isset($standard_options['process_once'])&&$standard_options['process_once']?'yes':'no'))) . '</dt>' . "\n";
    
    if (!empty($standard_options['fields'])) {
      $fields = array();
      foreach ($standard_options['fields'] as $name => $field) {
        if ($field['processed']) {
           $fields[] = $field['name'];
        }
      }
      if ($fields) {
        $output .= '<dt>' . t('Processed fields') . '</dt>' . "\n";
        $output .= '<dd>' . implode(', ', $fields) . '</dd>' . "\n";
      }
    }

    $output .= '</dl></dd>' . "\n";
  }
    
  if (!empty($options)) {
    $output .= '<dt>' . t('Service specific thread options') . '</dt>' . "\n";
    $output .= '<dd>' . "\n";
    $output .= render($options);
    $output .= '</dd>' . "\n";
  }

  $output .= '<dt>' . t('Configuration status') . '</dt>' . "\n";
  $output .= '<dd>' . "\n";
  $output .= theme('entity_status', array('status' => $status));
  $output .= '</dd>' . "\n";

  $output .= '</dl>';

  return $output;
}

/**
 * Form function for displaying a thread status form.
 *
 * @param ProcessApiThread $thread
 *   The thread whose status should be displayed.
 */
function process_api_admin_thread_status_form(array $form, array &$form_state, ProcessApiThread $thread) {
  $enabled = !empty($thread->enabled);
  $status = process_api_thread_status($thread);
  $server = $thread->server();

  $form['#attached']['css'][] = drupal_get_path('module', 'process_api') . '/process_api.admin.css';
  $form_state['thread'] = $thread;

  $form['status_message'] = array(
    '#type' => 'item',
    '#title' => t('Status'),
    '#description' => check_plain($enabled ? t('The thread is currently enabled.') : t('The thread is currently disabled.')),
  );
  if (!empty($server->enabled)) {
    $form['status'] = array(
      '#type' => 'submit',
      '#value' => $enabled ? t('Disable') : t('Enable'),
    );
  }

  $form['progress'] = array(
    '#type' => 'item',
    '#title' => t('Progress'),
  );
  $all = ($status['processed'] == $status['total']);
  if ($all) {
    $form['progress']['#description'] = t('All items have been processed (@total / @total).',
        array('@total' => $status['total']));
  }
  elseif (!$status['processed']) {
    $form['progress']['#description'] = t('All items still need to be processed (@total total).',
        array('@total' => $status['total']));
  }
  else {
    $percentage = (int) (100 * $status['processed'] / $status['total']);
    $form['progress']['#description'] = t('About @percentage% of all items have been processed in their latest version (@processed / @total).',
        array('@processed' => $status['processed'], '@total' => $status['total'], '@percentage' => $percentage));
  }

  if ($enabled && !$all) {
    $form['thread'] = array(
      '#type' => 'fieldset',
      '#title' => t('Process now'),
      '#attributes' => array('class' => array('container-inline')),
      '#collapsible' => TRUE,
    );
    $form['thread']['limit'] = array(
      '#type' => 'textfield',
      '#title' => t('Number of items:'),
      '#default_value' => empty($thread->options['cron_limit']) ? PROCESS_API_DEFAULT_CRON_LIMIT : $thread->options['cron_limit'],
      '#size' => 4,
      '#attributes' => array('class' => array('process-api-cron-limit')),
    );
    $form['thread']['button'] = array(
      '#type' => 'submit',
      '#value' => t('Process now'),
    );
    $form['thread']['total'] = array(
      '#type' => 'value',
      '#value' => $status['total'],
    );
    $form['thread']['remaining'] = array(
      '#type' => 'value',
      '#value' => $status['total'] - $status['processed'],
    );
  }

  if ($server) {
    if ($status['processed'] > 0) {
      $form['reprocess'] = array(
        '#type' => 'fieldset',
        '#title' => t('Re-processing'),
        '#collapsible' => TRUE,
      );
      $form['reprocess']['message'] = array(
        '#type' => 'item',
        '#description' => t('This will add all items to the thread again (overwriting the thread).'),
      );
      $form['reprocess']['button'] = array(
        '#type' => 'submit',
        '#value' => t('Re-process content'),
      );
    }

    $form['clear'] = array(
      '#type' => 'fieldset',
      '#title' => t('Clear thread'),
      '#collapsible' => TRUE,
    );
    $form['clear']['message'] = array(
      '#type' => 'item',
      '#description' => t('All items will be deleted from the thread and have to be inserted again by normally processing them.' .
          'Use with care, in most cases rebuilding the thread might be enough.'),
    );
    $form['clear']['button'] = array(
      '#type' => 'submit',
      '#value' => t('Clear thread'),
    );
  }

  return $form;
}

/**
 * Validation function for process_api_admin_thread_status_form.
 */
function process_api_admin_thread_status_form_validate(array $form, array &$form_state) {
  if ($form_state['values']['op'] == t('Process now') && !$form_state['values']['limit']) {
    form_set_error('number', t('You have to set the number of items to process. Set to -1 for processing all items.'));
  }
}

/**
 * Submit function for process_api_admin_thread_status_form.
 */
function process_api_admin_thread_status_form_submit(array $form, array &$form_state) {
  $redirect = &$form_state['redirect'];
  $values = $form_state['values'];
  $thread = $form_state['thread'];
  $pre = 'admin/config/system/process_api/thread/' . $thread->machine_name;
  switch ($values['op']) {
    case t('Enable'):
      $redirect = $pre . '/enable';
      break;
    case t('Disable'):
      $redirect = $pre . '/disable';
      break;
    case t('Process now'):
      $limit = $values['limit'];
      // @todo Batch API?
      try {
        $ret = process_api_process_items($thread, $limit);
      }
      catch (ProcessApiException $e) {
        $ret = FALSE;
        watchdog('process_api', 'An error occurred while trying to process items on thread !name: !message',
            array('!name' => $thread->machine_name, '!message' => $e->getMessage()), WATCHDOG_ERROR);
      }
      if ($ret) {
        drupal_set_message(format_plural($ret, 'Successfully processed 1 item.', 'Successfully processed @count items.'));
        if (($limit > 0 && $ret < min(array($limit, $values['remaining'])))
            || ($limit < 0 && $ret < $values['remaining'])) {
          drupal_set_message(t("Some items couldn't be processed. Check the logs for details."), 'warning');
        }
      }
      else {
        drupal_set_message(t("Couldn't process items. Check the logs for details."), 'error');
      }
      $redirect = $pre . '/status';
      break;
    case t('Re-process content'):
      $redirect = $pre . '/reprocess';
      break;
    case t('Clear thread'):
      $redirect = $pre . '/clear';
       break;

    default:
      throw new ProcessApiException(t('Unknown action.'));
  }
}

/**
 * Edit a thread' settings.
 *
 * @param ProcessApiThread $thread
 *   The thread to edit.
 */
function process_api_admin_thread_edit(array $form, array &$form_state, ProcessApiThread $thread) {

  $form_state['thread'] = $thread;

  $form['#attached']['css'][] = drupal_get_path('module', 'process_api') . '/process_api.admin.css';
  $form['name'] = array(
    '#type' => 'textfield',
    '#title' => t('Thread name'),
    '#maxlength' => 50,
    '#default_value' => $thread->name,
    '#required' => TRUE,
  );
  $form['description'] = array(
    '#type' => 'textarea',
    '#title' => t('Thread description'),
    '#default_value' => $thread->description,
  );
  
  $class = process_api_get_service_info($thread->ensureServer()->class);
  
  $form['server_dependent']['#collapsible'] = FALSE;
  $form['server_dependent']['#title'] = $class['name'];
  $form['server_dependent']['#description'] = $class['description'];
  $form['server_dependent']['#type'] = 'fieldset';    
  
  $options_form = $thread->configurationForm($form, $form_state);
  $form['server_dependent']['options']['form'] = $options_form ? $options_form : array('#markup' => t('There are no thread configuration options for this service class.'));
  $form['server_dependent']['options']['#tree'] = TRUE;

  $form['cron_limit'] = array(
    '#type' => 'textfield',
    '#title' => t('Cron limit'),
    '#description' => t('Set how many items will be processed at most during each run of cron. ' .
        '"0" means that no items will be processed by cron for this thread, "-1" means that cron should thread all items at once.'),
    '#default_value' => isset($thread->options['cron_limit']) ? $thread->options['cron_limit'] : PROCESS_API_DEFAULT_CRON_LIMIT,
    '#size' => 4,
    '#attributes' => array('class' => array('process-api-cron-limit')),
  );
  $form['live'] = array(
    '#type' => 'select',
    '#title' => t('Try process "live"'),
    '#options' => array('live' => t('Yes'), '' => t('No')),
    '#description' => t('Will try to process entities in live on insert and update. This may cost overhead.'),
    '#default_value' => isset($thread->options['live']) ? $thread->options['live'] : '',
  );

  $form['process_once'] = array(
      '#type' => 'select',
      '#title' => t('Process only once'),
      '#options' => array('process_once' => t('Yes'), '' => t('No')),
      '#description' => t('Will try to process entities only once at insert.'),
      '#default_value' => isset($thread->options['process_once']) ? $thread->options['process_once'] : '',
  );  

  $form['cascade_update'] = array(
      '#type' => 'select',
      '#title' => t('Cascade update'),
      '#options' => array('cascade_update' => t('Yes'), '' => t('No')),
      '#description' => t('Entity updates in this thread will trigger other threads.'),
      '#default_value' => isset($thread->options['cascade_update']) ? $thread->options['cascade_update'] : '',
  ); 
  
  $form['retroactive_process'] = array(
      '#type' => 'select',
      '#title' => t('Retroactive process'),
      '#options' => array('retroactive_process' => t('Yes'), '' => t('No')),
      '#description' => t('Process existing entities.'),
      '#default_value' => isset($thread->options['retroactive_process']) ? $thread->options['retroactive_process'] : '',
  );

  $form['track_disabled'] = array(
      '#type' => 'select',
      '#title' => t('Keep track when disabled'),
      '#options' => array('track_disabled' => t('Yes'), '' => t('No')),
      '#description' => t('Keep track of entities modifications even if thread is disabled.'),
      '#default_value' => isset($thread->options['track_disabled']) ? $thread->options['track_disabled'] : '',
  );  
  
  $form['purge_range'] = array(
      '#type' => 'textfield',
      '#title' => t('Purge range'),
      '#description' => t('For non retroactive process, you can define a purge of the Process Api tables. ' .
          'The purge range is in days. Items older than this range will be purged. ' .
          '"-1" means that purge is disabled.'),
      '#default_value' => isset($thread->options['purge_range']) ? $thread->options['purge_range'] : PROCESS_API_DEFAULT_PURGE_RANGE,
      '#size' => 4,
      '#attributes' => array('class' => array('process-api-purge-range')),
  );

  $form['node_operations'] = array(
      '#type' => 'fieldset',
      '#title' => t('Node operations'),
      '#description' => t('Live trigger Process API in bulk node operations.'),
      '#tree' => TRUE,
  );
  
  foreach (module_invoke_all('node_operations') as $operation => $array) {
    $form['node_operations'][$operation] = array(
        '#type' => 'select',
        '#title' => $array['label'],
        '#options' => array('live' => t('Yes'), '' => t('No')),
        '#default_value' => isset($thread->options['node_operations'][$operation]) ? $thread->options['node_operations'][$operation] : 'live',
    );
  }
  
  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Save settings'),
  );

  return $form;
}

/**
 * Validation callback for process_api_admin_thread_edit.
 */
function process_api_admin_thread_edit_validate(array $form, array &$form_state) {
  $cron_limit = $form_state['values']['cron_limit'];
  if ($cron_limit != '' . ((int) $cron_limit)) {
    // We don't enforce stricter rules and treat all negative values as -1.
    form_set_error('cron_limit', t('The cron limit must be a number.'));
  }
  $thread = $form_state['thread'];

  $thread->configurationFormValidate(
    $form,
    isset($form_state['values']['options']['form']) ? $form_state['values']['options']['form'] : array(),
    $form_state);

}

/**
 * Submit callback for process_api_admin_thread_edit.
 */
function process_api_admin_thread_edit_submit(array $form, array &$form_state) {
  form_state_values_clean($form_state);

  $values = $form_state['values'];
  $thread = $form_state['thread'];
  
  $thread->snapshot();
  if (empty($thread->options)) $thread->options = array();
  $thread->options['cron_limit']=$form_state['values']['cron_limit'];
  $thread->options['live']=$form_state['values']['live'];
  $thread->options['process_once']=$form_state['values']['process_once'];
  $thread->options['cascade_update']=$form_state['values']['cascade_update'];
  $thread->options['retroactive_process']=$form_state['values']['retroactive_process'];
  $thread->options['track_disabled']=$form_state['values']['track_disabled'];
  $thread->options['purge_range']=$form_state['values']['purge_range'];
  $thread->options['node_operations']=$form_state['values']['node_operations'];
  
  $thread->configurationFormSubmit(
    $form,
    isset($form_state['values']['options']['form'])?$form_state['values']['options']['form']:array(),
    $form_state);
  
  unset($values['options']);
  unset($values['cron_limit']);
  unset($values['live']);
  unset($values['process_once']);
  unset($values['cascade_update']);
  unset($values['retroactive_process']);
  unset($values['track_disabled']);
  unset($values['purge_range']);
  unset($values['node_operations']);
  $thread->setValues($values);

  $ret = $thread->update();
  $form_state['redirect'] = 'admin/config/system/process_api/thread/' . $thread->machine_name;
  if ($ret) {
    drupal_set_message(t('The process thread was successfully edited.'));
  }
  else {
    drupal_set_message(t('No values were changed.'));
  }
}

/**
 * Edit a thread' workflow (data alter callbacks, pre-/postprocessors, and their
 * order).
 *
 * @param ProcessApiThread $thread
 *   The thread to edit.
 */
// Copied from filter_admin_format_form
function process_api_admin_thread_workflow(array $form, array &$form_state, ProcessApiThread $thread) {
  $callback_info = process_api_get_alter_callbacks();
  $processor_info = process_api_get_processors();
  $options = empty($thread->options) ? array() : $thread->options;

  $form_state['thread'] = $thread;
  $form['#tree'] = TRUE;
  $form['#attached']['js'][] = drupal_get_path('module', 'process_api') . '/process_api.admin.js';

  // Callbacks

  $callbacks = empty($options['data_alter_callbacks']) ? array() : $options['data_alter_callbacks'];
  foreach ($callback_info as $name => $callback) {
    if (!isset($callbacks[$name])) {
      $callbacks[$name]['status'] = 0;
      $callbacks[$name]['weight'] = $callback['weight'];
    }
  }
  $form['#callbacks'] = $callbacks;
  $form['callbacks'] = array(
    '#type' => 'fieldset',
    '#title' => t('Data alterations'),
    '#description' => t('Select the alterations that will be executed on processed items, and their order.'),
    '#collapsible' => TRUE,
  );

  // Callback status.
  $form['callbacks']['status'] = array(
    '#type' => 'item',
    '#title' => t('Enabled data alterations'),
    '#prefix' => '<div class="process-api-status-wrapper">',
    '#suffix' => '</div>',
  );
  foreach ($callback_info as $name => $callback) {
    $form['callbacks']['status'][$name] = array(
      '#type' => 'checkbox',
      '#title' => $callback['name'],
      '#default_value' => $callbacks[$name]['status'],
      '#parents' => array('callbacks', $name, 'status'),
      '#description' => $callback['description'],
      '#weight' => $callback['weight'],
    );
  }

  // Callback order (tabledrag).
  $form['callbacks']['order'] = array(
    '#type' => 'item',
    '#title' => t('Data alteration processing order'),
    '#theme' => 'process_api_admin_item_order',
    '#table_id' => 'process-api-callbacks-order-table',
  );
  foreach ($callback_info as $name => $callback) {
    $form['callbacks']['order'][$name]['item'] = array(
      '#markup' => $callback['name'],
    );
    $form['callbacks']['order'][$name]['weight'] = array(
      '#type' => 'weight',
      '#delta' => 50,
      '#default_value' => $callbacks[$name]['weight'],
      '#parents' => array('callbacks', $name, 'weight'),
    );
    $form['callbacks']['order'][$name]['#weight'] = $callbacks[$name]['weight'];
  }

  // Processors

  $processors = empty($options['processors']) ? array() : $options['processors'];
  $processor_objects = array();
  foreach ($processor_info as $name => $processor) {
    if (!isset($processors[$name])) {
      $processors[$name]['status'] = 0;
      $processors[$name]['weight'] = $processor['weight'];
    }
    $settings = empty($processors[$name]['settings']) ? array() : $processors[$name]['settings'];
    if (class_exists($processor['class'])) {
      $processor_objects[$name] = new $processor['class']($thread, $settings);
    }
    if (!(class_exists($processor['class']) && $processor_objects[$name] instanceof ProcessApiProcessorInterface)) {
      watchdog('process api', 'Processor !id specifies illegal processor class !class.', array('!id' => $name, '!class' => $processor['class']), WATCHDOG_WARNING);
      unset($processor_info[$name]);
      unset($processors[$name]);
      unset($processor_objects[$name]);
      continue;
    }
  }
  $form['#processors'] = $processors;
  $form['processors'] = array(
    '#type' => 'fieldset',
    '#title' => t('Processors'),
    '#description' => t('Select processors which will pre- and post-process data at process time, and their order. ' .
        'Most processors will only influence fulltext fields, but refer to their individual descriptions for details regarding their effect.'),
    '#collapsible' => TRUE,
  );

  // Processor status.
  $form['processors']['status'] = array(
    '#type' => 'item',
    '#title' => t('Enabled processors'),
    '#prefix' => '<div class="process-api-status-wrapper">',
    '#suffix' => '</div>',
  );
  foreach ($processor_info as $name => $processor) {
    $form['processors']['status'][$name] = array(
      '#type' => 'checkbox',
      '#title' => $processor['name'],
      '#default_value' => $processors[$name]['status'],
      '#parents' => array('processors', $name, 'status'),
      '#description' => $processor['description'],
      '#weight' => $processor['weight'],
    );
  }

  // Processor order (tabledrag).
  $form['processors']['order'] = array(
    '#type' => 'item',
    '#title' => t('Processor processing order'),
    '#description' => t('Set the order in which preprocessing will be done at process time.'),
    '#theme' => 'process_api_admin_item_order',
    '#table_id' => 'process-api-processors-order-table',
  );
  foreach ($processor_info as $name => $processor) {
    $form['processors']['order'][$name]['item'] = array(
      '#markup' => $processor['name'],
    );
    $form['processors']['order'][$name]['weight'] = array(
      '#type' => 'weight',
      '#delta' => 50,
      '#default_value' => $processors[$name]['weight'],
      '#parents' => array('processors', $name, 'weight'),
    );
    $form['processors']['order'][$name]['#weight'] = $processors[$name]['weight'];
  }

  // Processor settings.
  $form['processors']['settings_title'] = array(
    '#type' => 'item',
    '#title' => t('Processor settings'),
  );
  $form['processors']['settings'] = array(
    '#type' => 'vertical_tabs',
  );

  foreach ($processor_info as $name => $processor) {
    $settings_form = $processor_objects[$name]->configurationForm();
    if (!empty($settings_form)) {
      $form['processors']['settings'][$name] = array(
        '#type' => 'fieldset',
        '#title' => $processor['name'],
        '#parents' => array('processors', $name, 'settings'),
        '#weight' => $processor['weight'],
        '#group' => 'processors_settings',
      );
      $form['processors']['settings'][$name] += $settings_form;
    }
  }

  $form['actions'] = array('#type' => 'actions');
  $form['actions']['submit'] = array('#type' => 'submit', '#value' => t('Save configuration'));

  return $form;
}

/**
 * Returns HTML for a processor/callback order form.
 *
 * @param array $variables
 *   An associative array containing:
 *   - element: A render element representing the form.
 */
function theme_process_api_admin_item_order(array $variables) {
  $element = $variables['element'];

  $rows = array();
  foreach (element_children($element, TRUE) as $name) {
    $element[$name]['weight']['#attributes']['class'][] = 'process-api-order-weight';
    $rows[] = array(
      'data' => array(
        drupal_render($element[$name]['item']),
        drupal_render($element[$name]['weight']),
      ),
      'class' => array('draggable'),
    );
  }
  $output = drupal_render_children($element);
  $output .= theme('table', array('rows' => $rows, 'attributes' => array('id' => $element['#table_id'])));
  drupal_add_tabledrag($element['#table_id'], 'order', 'sibling', 'process-api-order-weight', NULL, NULL, TRUE);

  return $output;
}

/**
 * Submit callback for process_api_admin_thread_workflow.
 */
function process_api_admin_thread_workflow_submit(array $form, array &$form_state) {
  $values = $form_state['values'];
  unset($values['processors']['settings']);
  $thread = $form_state['thread'];

  $options = empty($thread->options) ? array() : $thread->options;
  $fields_set = !empty($options['fields']);

  // Set default processed status for newly added fields.
  foreach ($values['callbacks'] as $func => $callback) {
    if ($callback['status'] && empty($options['callbacks'][$func]['status'])) {
      if (!is_callable($func)) {
        watchdog('process_api', 'Undefined data alter callback function !function() specified in thread !name', array('!function' => $func, '!name' => $thread->name), WATCHDOG_WARNING);
        continue;
      }
      $items = array();
      $properties = $func($thread, $items);
      $types = process_api_field_types();
      if ($properties) {
        foreach ($properties as $key => $field) {
          $type = $field['type'];
          $inner = process_api_extract_inner_type($type);
          if ($inner != 'token' && empty($types[$inner])) {
            // Someone apparently added a structure or entity as a property in a data-alter callback.
            continue;
          }
          if ($inner == 'token' || (process_api_is_text_type($inner) && !empty($field['options list']))) {
            $old = $type;
            $type = 'string';
            while (process_api_is_list_type($old)) {
              $old = substr($old, 5, -1);
              $type = "list<$type>";
            }
          }
          $options['fields'][$key] = array(
            'name' => $field['label'],
            'type' => $type,
            'boost' => '1.0',
            'processed' => 1,
          );
        }
      }
    }
  }

  $options['data_alter_callbacks'] = $values['callbacks'];
  $options['processors'] = $values['processors'];

  // Save the already sorted arrays to avoid having to sort them at each use.
  uasort($options['data_alter_callbacks'], 'process_api_admin_element_compare');
  uasort($options['processors'], 'process_api_admin_element_compare');

  $thread->snapshot('options');
  $thread->options=$options;
  
  $ret = $thread->update();
  if ($ret) {
    $thread->reprocess();
    drupal_set_message(t("The process thread' workflow was successfully edited. " .
        'All content was scheduled for re-processing so the new settings can take effect.'));
  }
  else {
    drupal_set_message(t('No values were changed.'));
  }

  if (!$fields_set) {
    drupal_set_message(t('Please set the fields that should be processed.'));
    $form_state['redirect'] = 'admin/config/system/process_api/thread/' . $thread->machine_name . '/fields';
  }
  else {
    $form_state['redirect'] = 'admin/config/system/process_api/thread/' . $thread->machine_name;
  }
}

/**
 * Sort callback sorting array elements by their "weight" key, if present.
 *
 * @see element_sort
 */
function process_api_admin_element_compare($a, $b) {
  $a_weight = (is_array($a) && isset($a['weight'])) ? $a['weight'] : 0;
  $b_weight = (is_array($b) && isset($b['weight'])) ? $b['weight'] : 0;
  if ($a_weight == $b_weight) {
    return 0;
  }
  return ($a_weight < $b_weight) ? -1 : 1;
}

/**
 * Select the processed fields.
 *
 * @param ProcessApiThread $thread
 *   The thread to edit.
 */
function process_api_admin_thread_fields(array $form, array &$form_state, ProcessApiThread $thread) {

  $options = process_api_get_fields($thread);

  $fields = $options['fields'];
  $additional = $options['additional_structures'];

  // An array of option arrays for types, keyed by nesting level.
  $types = array(0 => process_api_field_types());
  $boosts = drupal_map_assoc(array('0.1', '0.2', '0.3', '0.5', '0.8', '1.0', '2.0', '3.0', '5.0', '8.0', '13.0', '21.0'));

  $form_state['thread'] = $thread;
  
  $form['#theme'] = 'process_api_admin_fields_table';
  $form['#thread'] = $thread;
  $form['#tree'] = TRUE;
  $form['description'] = array(
    '#type' => 'item',
    '#title' => t('Select fields to process'),
    '#description' => t('<p>Select the fields you want to pass to the thread.</p>'),
  );
  if ($thread->server) {
    $form['description']['#description'] .= '<p>' . t('Check the <a href="@server-url">' . "server's</a> service class description for details.",
        array('@server-url' => url('admin/config/system/process_api/server/' . $thread->server))) . '</p>';
  }
  foreach ($fields as $key => $info) {
    $form['fields'][$key]['title']['#markup'] = check_plain($info['name']);
    $form['fields'][$key]['name'] = array(
      '#type' => 'value',
      '#value' => $info['name'],
    );
    if (isset($info['description'])) {
      $form['fields'][$key]['description'] = array(
        '#type' => 'value',
        '#value' => $info['description'],
      );
    }
    // Determine the correct type options (i.e., with the correct nesting level).
    $level = process_api_list_nesting_level($info['type']);
    if (empty($types[$level])) {
      $type_prefix = str_repeat('list<', $level);
      $type_suffix = str_repeat('>', $level);
      $types[$level] = array();
      foreach ($types[0] as $type => $name) {
        // We use the singular name for list types, since the user usually doesn't care about the nesting level.
        $types[$level][$type_prefix . $type . $type_suffix] = $name;
      }
    }

    $form['fields'][$key]['type'] = array(
      '#type' => 'value',
      '#value' => $info['type'],
    );

    $form['fields'][$key]['processed'] = array(
      '#type' => 'checkbox',
      '#default_value' => $info['processed'],
    );
    
    // custom inputs
    foreach ($thread->fieldsConfigurationHeaders() as $header_id => $label) {
      $form['fields'][$key][$header_id]=$thread->fieldConfigurationForm($header_id, $info, $form, $form_state);
    }
    
    if ($key == 'process_api_language') {
      // Is treated specially to always thread the language.
      $form['fields'][$key]['processed']['#default_value'] = 1;
      $form['fields'][$key]['processed']['#disabled'] = TRUE;
    }
  }

  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Save changes'),
  );

  if ($additional) {
    reset($additional);
    $form['additional'] = array(
      '#type' => 'fieldset',
      '#title' => t('Add related entity'),
      '#description' => t('There are entities related to entities of this type. ' .
          'You can add their fields to the list above so they can be processed too.') . '<br />',
      '#collapsible' => TRUE,
      '#collapsed' => TRUE,
      '#attributes' => array('class' => array('container-inline')),
      'field' => array(
        '#type' => 'select',
        '#options' => array_merge(array('' => t('(choose)')), $additional),
        '#default_value' => '',
      ),
      'add' => array(
        '#type' => 'submit',
        '#value' => t('Add fields'),
      )
    );
  }

  if (isset($options['structures'])&&!empty($options['structures'])) {
    $form['structures'] = array(
      '#type' => 'fieldset',
      '#title' => t('Added entities'),
      '#tree' => TRUE,
      '#collapsible' => TRUE,
      '#collapsed' => TRUE,
    );
    foreach ($options['structures'] as $prefix => $name) {
      if (!$prefix) continue;
      $form['structures']['list'][$prefix]['delete'] = array(
        '#type' => 'checkbox',
        '#default_value' => FALSE,
        '#title' => t('Delete @struct', array('@struct' => check_plain($name))),
      );
    }
    
    $form['structures']['remove'] = array(
        '#type' => 'submit',
        '#value' => t('Remove fields'),
    );
  }  
  
  return $form;
  // @todo Provide a way to remove a related entity's fields again?
}


/**
 * Returns HTML for a field list form.
 *
 * @param array $variables
 *   An associative array containing:
 *   - element: A render element representing the form.
 */
function theme_process_api_admin_fields_table($variables) {
  $form = $variables['element'];
  $thread = $form['#thread']; //TODO ugly but 'render element' => 'element' is black box for me

  $header = array(t('Field'), t('Processed'));
  foreach ($thread->fieldsConfigurationHeaders() as $header_id => $label) {
    $header[]=$label;
  }

  $rows = array();
  foreach (element_children($form['fields']) as $name) {
    $row = array();
    foreach (element_children($form['fields'][$name]) as $field) {
      if ($cell = render($form['fields'][$name][$field])) {
        $row[] = $cell;
      }
    }
    
    if (empty($form['fields'][$name]['description']['#value'])) {
      $rows[] = $row;
    }
    else {
      $rows[] = array(
        'data' => $row,
        'title' => $form['fields'][$name]['description']['#value'],
      );
    }
  }

  // TODO omg use weight !!!
  $submit = $form['submit'];
  $additional = isset($form['additional']) ? $form['additional'] : FALSE;
  $structures = isset($form['structures']) ? $form['structures'] : FALSE;
  unset($form['submit'], $form['additional'], $form['structures']);
  $output = drupal_render_children($form);
  $output .= theme('table', array('header' => $header, 'rows' => $rows));
  $output .= render($submit);
  if ($additional) {
    $output .= render($additional);
  }
  if ($structures) {
    $output .= render($structures);
  }

  return $output;
}

/**
 * Submit function for process_api_admin_thread_fields.
 */
function process_api_admin_thread_fields_submit(array $form, array &$form_state) {
  $thread = $form_state['thread'];
  //$options = isset($thread->options) ? $thread->options : array();
  if ($form_state['values']['op'] == t('Save changes')) {
    $fields = $form_state['values']['fields'];
    foreach ($fields as $name => &$field) {
      unset($field['description']);
    }
    $thread->snapshot('options');
    $thread->options['fields']=$fields;
    $ret = $thread->update();

    //$form_state['redirect'] = 'admin/config/system/process_api/thread/' . $thread->machine_name;
    if ($ret) {
      drupal_set_message(t('The processed fields were successfully changed. ' .
          'The thread was not cleared.'));
    }
    else {
      drupal_set_message(t('No values were changed.'));
    }
    return;
  }

  $structures = isset($thread->options['structures']) ? $thread->options['structures'] : array();
  if ($form_state['values']['additional']['field']) {  
    $structures[$form_state['values']['additional']['field']] = $form['additional']['field']['#options'][$form_state['values']['additional']['field']];
  }
  $to_del = array();
  
  // user wants some structure to be removed
  if (isset($form_state['values']['structures']['list'])) {
    foreach($form_state['values']['structures']['list'] as $prefix => $info) {
      if ($info['delete']) {
        unset($structures[$prefix]);
        // stack prefix to remove
        $to_del[$prefix]=$prefix;
      }
    }
  }
  
  // snapshot so we can detect critical modifications (trigger in update()) 
  $thread->snapshot('options');
  
  if (!empty($to_del)) {
    $fields=array_keys($thread->options['fields']);
    foreach ($fields as $field) {
      $i = strrpos($field, ':');
      if ($i===FALSE) continue;
      $prefix = substr($field, 0, $i);
      if (isset($to_del[$prefix])&&isset($thread->options['fields'][$field])) {
        // remove fields of the unwanted structure
        unset($thread->options['fields'][$field]);
      }
    }
  }
  
  $thread->options['structures']=$structures;
  $ret = $thread->update();

  if ($ret) {
    drupal_set_message(t('The available fields were successfully changed.'));
  }
  else {
    drupal_set_message(t('No values were changed.'));
  }

  $form_state['redirect'] = 'admin/config/system/process_api/thread/' . $thread->machine_name . '/fields';
}


/**
 * Helper function for displaying a generic confirmation form.
 *
 * @return
 *   Either a form array, or FALSE if this combination of type and action is
 *   not supported.
 */
function process_api_admin_confirm(array $form, array &$form_state, $type, $action, Entity $entity) {
  switch ($type) {
    case 'server':
      switch ($action) {
        case 'disable':
          $text = array(
            t('Disable server @name', array('@name' => $entity->name)),
            t('Do you really want to disable this server?'),
            t('This will disable both the server and all associated threads.'),
            t('The server and its threads were successfully disabled.'),
          );
          break;
        case 'delete':
          if ($entity->status == ENTITY_OVERRIDDEN) {
            $text = array(
              t('Revert server @name', array('@name' => $entity->name)),
              t('Do you really want to revert this server?'),
              t('This will revert all settings for this server back to the defaults. This action cannot be undone.'),
              t('The server settings have been successfully reverted.'),
            );
          }
          else {
            $text = array(
              t('Delete server @name', array('@name' => $entity->name)),
              t('Do you really want to delete this server?'),
              t('This will delete the server and all associated threads.'),
              t('The server was successfully deleted.'),
            );
          }
          break;
        default:
          return FALSE;
      }
      break;
    case 'thread':
      switch ($action) {
        case 'reprocess':
          $text = array(
            t('Re-process thread @name', array('@name' => $entity->name)),
            t('Do you really want to mark this thread to re-process ?'),
            t("All items will be re-processed."),
            t('The thread was successfully marked to re-process.'),
          );
          break;
        case 'clear':
          $text = array(
            t('Clear thread @name', array('@name' => $entity->name)),
            t('Do you really want to clear this thread ?'),
            t("All items will be cleared."),
            t('The thread was successfully cleared.'),
          );
          break;          
        case 'disable':
          $text = array(
            t('Disable thread @name', array('@name' => $entity->name)),
            t('Do you really want to disable this thread?'),
            t("Process on this thread won't be available until it is re-enabled."),
            t('The thread was successfully disabled.'),
          );
          break;
        case 'delete':
          if ($entity->status == ENTITY_OVERRIDDEN) {
            $text = array(
              t('Revert thread @name', array('@name' => $entity->name)),
              t('Do you really want to revert this thread?'),
              t('This will revert all settings on this thread back to the defaults. This action cannot be undone.'),
              t('The thread settings have been successfully reverted.'),
            );
          }
          else {
            $text = array(
              t('Delete thread @name', array('@name' => $entity->name)),
              t('Do you really want to delete this thread?'),
              t('This will remove the thread from the server and delete all settings. ' .
                  'All data on this thread will be lost.'),
              t('The thread has been successfully deleted.'),
            );
          }
          break;
        default:
          return FALSE;
      }
      break;
    default:
      return FALSE;
  }

  $form = array(
    'type' => array(
      '#type' => 'value',
      '#value' => $type,
    ),
    'action' => array(
      '#type' => 'value',
      '#value' => $action,
    ),
    'id' => array(
      '#type' => 'value',
      '#value' => $entity->machine_name,
    ),
    'message' => array(
      '#type' => 'value',
      '#value' => $text[3],
    ),
  );
  $desc = "<h3>{$text[1]}</h3><p>{$text[2]}</p>";
  return confirm_form($form, $text[0], "admin/config/system/process_api/$type/{$entity->machine_name}", filter_xss($desc));
}

/**
 * Submit function for process_api_admin_confirm().
 */
function process_api_admin_confirm_submit(array $form, array &$form_state) {
  $values = $form_state['values'];

  $type = $values['type'];
  $action = $values['action'];
  $id = $values['id'];

  $function = "process_api_{$type}_{$action}";
  if ($function($id)) {
    drupal_set_message($values['message']);
  }
  else {
    drupal_set_message(t('An error has occurred while performing the desired action. Check the logs for details.'), 'error');
  }

  $form_state['redirect'] = $action == 'delete'
      ? ( $type == 'thread' ? "admin/config/system/process_api/threads" : "admin/config/system/process_api" )
      : "admin/config/system/process_api/$type/$id";
}

/**
 * Delete a field from a server configuration
 *
 * @param ProcessApiThread $thread
 * @param string $field_name Name of the field to delete
 */
function process_api_admin_field_delete(ProcessApiThread $thread, $field_name) {
  // refresh the fields list for the current thread

  if(isset($thread->options['fields'][$field_name])) {

    if( count(preg_split('/:/', $field_name)) == 1) {
      drupal_set_message('You can\'t delete this native field', 'error');
    }
    else{
      unset($thread->options['fields'][$field_name]);
      $ret =$thread->save();
      if(!$ret) drupal_set_message('The field can\'t be deleted', 'error');
      else drupal_set_message('The field has been deleted.');
    }
  }
  else drupal_set_message('This field doesn\'t exist', 'error');

  drupal_goto('admin/config/system/process_api/thread/' . $thread->machine_name . '/fields');
}

/**
 * Form builder; process api admin settings form.
 *
 * @ingroup forms
 * @see system_settings_form()
 */
function process_api_admin_settings($form, &$form_state) {

  $form['process_api_node_operations_form_alter'] = array(
      '#type' => 'textarea',
      '#title' => t("Nodes admin form for bulk nodes operations"),
      '#default_value' => variable_get('process_api_node_operations_form_alter', 'node_admin_nodes'),
      '#description' => t("One form per line."),
  );

  $form = system_settings_form($form);

  return $form;
}