<?php

/**
 * Implementation of hook_drush_command().
 */
function process_api_drush_command() {
  $items = array();

  $items['process-api-process'] = array(
    'description' => "Do the process.",
    'arguments' => array(
      'thread_machine_name' => 'Thread machine name(s)',
    ),
    'options' => array(
      'max' => 'Max number of items to process. Default is -1 for no max.',
    ),
    'examples' => array(
      'drush pap --max=50 my_thread1,my_thread2' => 'Process my thread 1 and 2 with a maximum of 50 items.',
    ),
    'aliases' => array('pap'),
    'bootstrap' => DRUSH_BOOTSTRAP_DRUPAL_FULL,
  );

  $items['process-api-process-all'] = array(
    'description' => "Do the process.",
    'arguments' => array(
      ),
    'options' => array(
       'max' => 'Max number of items to process. Default is -1 for no max.',
      ),
    'examples' => array(
      'drush papa --max=50' => 'Process all thread with a maximum of 50 items.',
      ),
    'aliases' => array('papa'),
        'bootstrap' => DRUSH_BOOTSTRAP_DRUPAL_FULL,
      );
  
  $items['process-api-purge'] = array(
      'description' => "Do the purge.",
      'arguments' => array(
          'thread_machine_name' => 'Thread machine name(s)',
      ),
      'options' => array(
          'range' => 'Purge range in days. Default is to take range defined in settings.',
      ),
      'examples' => array(
          'drush papurge --range=50 my_thread1,my_thread2' => 'Purge my thread 1 and 2 for items modified before the past 50 days.',
      ),
      'aliases' => array('papurge'),
      'bootstrap' => DRUSH_BOOTSTRAP_DRUPAL_FULL,
  );
  
  $items['process-api-purge-all'] = array(
      'description' => "Do the purge.",
      'arguments' => array(
      ),
      'options' => array(
          'range' => 'Purge range in days. Default is to take range defined in settings for each thread.',
      ),
      'examples' => array(
          'drush papurgeall --range=50' => 'Purge all thread with a for items modified before the past 50 days.',
      ),
      'aliases' => array('papurgeall'),
      'bootstrap' => DRUSH_BOOTSTRAP_DRUPAL_FULL,
  );  
    return $items;
}

/**
 * Implementation of hook_drush_help().
 */
function process_api_drush_help($section) {
  switch ($section) {
    case 'drush:process-api-process':
      return dt("This command will process the given thread.");
    case 'meta:process-api-process:title':
      return dt("Process API commands");
    case 'meta:process-api-process:summary':
      return dt("Help you process threads.");
  }
}

/**
 * Implementation of drush_hook_COMMAND().
 */
function drush_process_api_process($thread_machine_names) {
  global $user;
  if ($account = user_load(1)) {
    $user = $account;
  }
  foreach (explode(',', $thread_machine_names) as $thread_machine_name) {
    $thread_machine_name = trim($thread_machine_name);
    $thread = process_api_thread_load_by_machine_name($thread_machine_name);
    drush_print("processing {$thread->name}...");
    
    if (!$thread) {
      drush_set_error(t("Cannot load thread !thread", array('!thread' => $thread_machine_name)));
      drush_die();
    }
    
    try {
      $num = process_api_process_items($thread, drush_get_option('max', -1));
      if ($num) {
        watchdog('process_api', 'Processed !num items for thread !name', array('!num' => $num, '!name' => $thread->name), WATCHDOG_INFO);
      }
    }
    catch (ProcessApiException $e) {
      watchdog('process_api', $e->getMessage(), NULL, WATCHDOG_WARNING);
    }
  }
}

/**
 * Implementation of drush_hook_COMMAND().
 */
function drush_process_api_purge($thread_machine_names) {
  global $user;
  if ($account = user_load(1)) {
    $user = $account;
  }
  foreach (explode(',', $thread_machine_names) as $thread_machine_name) {
    $thread_machine_name = trim($thread_machine_name);
    $thread = process_api_thread_load_by_machine_name($thread_machine_name);
    drush_print("purging {$thread->name}...");
  
    if (!$thread) {
      drush_set_error(t("Cannot load thread !thread", array('!thread' => $thread_machine_name)));
      drush_die();
    }
  
    try {
      $num = process_api_purge_items($thread, drush_get_option('range', -1));
      if ($num) {
        watchdog('process_api', 'Purged items for thread !name', array('!name' => $thread->name), WATCHDOG_INFO);
      }
    }
    catch (ProcessApiException $e) {
      watchdog('process_api', $e->getMessage(), NULL, WATCHDOG_WARNING);
    }
  }
}

/**
* Implementation of drush_hook_COMMAND().
*/
function drush_process_api_purge_all() {
  global $user;
  if ($account = user_load(1)) {
    $user = $account;
  }
  $range = drush_get_option('range', -1);
  foreach (process_api_get_enabled_threads() as $thread) {
    drush_print("purging {$thread->name}...");
    try {
      $num = process_api_purge_items($thread, $range);
      if ($num) {
        watchdog('process_api', 'Purged items for thread !name', array('!name' => $thread->name), WATCHDOG_INFO);
      }
    }
    catch (ProcessApiException $e) {
      watchdog('process_api', $e->getMessage(), NULL, WATCHDOG_WARNING);
    }
  }
}

/**
 * Implementation of drush_hook_COMMAND().
 */
function drush_process_api_process_all() {
  global $user;
  if ($account = user_load(1)) {
    $user = $account;
  }
  $limit = drush_get_option('max', -1);
  foreach (process_api_get_enabled_threads() as $thread) {
    drush_print("processing {$thread->name}...");
    try {
      $num = process_api_process_items($thread, $limit);
      if ($num) {
        watchdog('process_api', 'Processed !num items for thread !name', array('!num' => $num, '!name' => $thread->name), WATCHDOG_INFO);
      }
    }
    catch (ProcessApiException $e) {
      watchdog('process_api', $e->getMessage(), NULL, WATCHDOG_WARNING);
    }
  }
}